/*
 * Copyright © 2017-2024 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of Nvidia Corporation and its affiliates
 * (the "Company") and all right, title, and interest in and to the software
 * product, including all associated intellectual property rights, are and
 * shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 */

#define _GNU_SOURCE  // for asprintf
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "api/clx_api.h"

//
// The application's configuration
//
const char* kSourceId                 = "example";
const char* kSourceTag                = "generic";
const uint16_t kNumberOfCounters      = 10;
const uint16_t kNumberOfCounterGroups = 2;
const uint16_t kNumberOfSamples       = 200;
const uint32_t kSamplingPeriod        = 1000;
const uint32_t kForcePageWritePeriod  = 10 * 1000;

const char* kDataRootConfigurationVariable      = "DATA_ROOT";
const char* kExportersPathConfigurationVariable = "CLX_EXPORTERS_PATH";

//
// CLX API parameters
//
bool set_example_parameters(clx_api_params_t* params) {
    // Self identifiers
    params->source_id = strdup(kSourceId);
    if (params->source_id == NULL) {
        return false;
    }

    params->source_tag = strdup(kSourceTag);
    if (params->source_tag == NULL) {
        return false;
    }

    // Data collection parameters

    // for counters set buffer size to zero to send data immediately
    // or make it large (e.g 60000) to reduce num of transactions
    params->buffer_size = 1024;

    params->enable_opaque_events = false;

    // Netflow (exporter or provider?)
    params->netflow_collector_addr  = NULL;
    params->netflow_collector_port  = 0;
    params->netflow_max_packet_size = 0;

    // Data exporters:
    // ...plugins
    const char* exporters_path = getenv(kExportersPathConfigurationVariable);
    if (exporters_path != NULL) {
        params->exporters_dir = strdup(exporters_path);
        if (params->exporters_dir == NULL) {
            return false;
        }
    }

    // ...writer
    const char* data_root = getenv(kDataRootConfigurationVariable);
    if (data_root != NULL) {
        params->file_write_enabled = true;
        params->data_root          = strdup(data_root);
        params->schema_path        = strdup(data_root);
        params->data_path_template = strdup(CLX_API_DATA_PATH_TEMPLATE);
        params->max_file_size      = 32 * 1024;          // 32 kB
        params->max_file_age       = 2 * 60 * 1000000L;  // 2 mins

        if (params->data_root == NULL || params->schema_path == NULL || params->data_path_template == NULL) {
            return false;
        }
    }

    // ...IPC
    params->ipc_enabled                = false;
    params->ipc_sockets_dir            = NULL;
    params->ipc_max_reattach_time_msec = 1000;
    params->ipc_max_reattach_tries     = 10;
    params->ipc_socket_timeout_msec    = 3000;

    // HTTP (Prometheus) endpoint
    params->prometheus_endpoint = NULL;
    params->cset_dir_path       = NULL;

    return true;
}

void print_example_parameters(const clx_api_params_t* params, FILE* out) {
    fprintf(out, "Configuration parameters:\n");
    fprintf(out, "  Exporters\n");
    fprintf(out, "    exporters_dir      : %s\n", params->exporters_dir);
    fprintf(out, "  Data writer\n");
    fprintf(out, "    buffer_size        : %zu\n", params->buffer_size);
    fprintf(out, "    buffer_size        : %zu\n", params->buffer_size);
    fprintf(out, "    file_write_enabled : %s\n", params->file_write_enabled ? "true" : "false");
    fprintf(out, "    data_root          : %s\n", params->data_root);
    fprintf(out, "    schema_path        : %s\n", params->schema_path);
}

//
// Data provider with its internal configuration, state and methods
//
typedef struct example_provider_context_t {
    uint16_t num_groups;           // total number of counter groups
    uint16_t num_counters;         // total number of counters per group
    uint64_t iteration;            // sampling iteration, as source of the counter values
    uint32_t first_counter_idx[];  // index of the first counter in the group
} example_provider_context_t;

example_provider_context_t* construct_example_provider_context(uint16_t num_groups, uint16_t num_counters) {
    size_t size                     = sizeof(example_provider_context_t) + sizeof(uint32_t) * num_groups;
    example_provider_context_t* ret = (example_provider_context_t*)calloc(1, size);
    if (ret == NULL) {
        return NULL;
    }
    ret->num_groups   = num_groups;
    ret->num_counters = num_counters;
    return ret;
}

void destroy_example_provider_context(example_provider_context_t* provider_context) { free(provider_context); }

uint64_t get_counter_value(example_provider_context_t* provider_context, uint32_t group_num, uint32_t counter_num) {
    return group_num * 100000 + counter_num * 1000 + provider_context->iteration;
}

bool sample_counters(example_provider_context_t* provider_context, clx_api_context_t* ctx) {
    uint32_t data_size = 0;
    uint64_t timestamp = clx_api_get_timestamp();

    // get the buffer area to write counter data to
    uint64_t* data = (uint64_t*)clx_api_get_counters_buffer(ctx, timestamp, &data_size);
    if (data == NULL) {
        puts("Error: failed to get buffer");
        return false;
    }

    printf("    writing counters to the buffer %p of size %u\n", data, data_size);

    // "reading" the counter values
    uint64_t idx = 0;
    for (uint16_t g = 0; g < provider_context->num_groups; g++) {
        idx += provider_context->first_counter_idx[g];
        for (uint16_t n = 0; n < provider_context->num_counters; n++) {
            data[idx++] = get_counter_value(provider_context, g, n);
        }
    }

    provider_context->iteration++;

    return true;
}

bool write_bulk_data(example_provider_context_t* provider_context, clx_api_context_t* api_context) {
    // Construct some meaningful message
    char now_string[26];
    struct tm tm_info;
    struct timespec ts;
    clock_gettime(CLOCK_REALTIME, &ts);
    localtime_r(&ts.tv_sec, &tm_info);
    strftime(now_string, sizeof(now_string), "%Y-%m-%d %H:%M:%S", &tm_info);

    char* payload = NULL;
    if (asprintf(&payload, "[%s] CLX API example has sampled counters %zu times", now_string, provider_context->iteration) < 0) {
        puts("Error: failed to construct a message to post");
        return false;
    }

    printf("    posting message: \"%s\"\n", payload);
    bool ok = clx_api_write_blob_data(api_context, payload, strlen(payload));
    free(payload);
    return ok;
}

//
// CLX API interface object
//
bool setup_example_group_counters(clx_api_context_t* ctx, const char* group_name, uint16_t num_counters,
                                  uint32_t* first_counter_idx) {
    clx_api_counter_info_t counter = {.counter_name  = NULL,
                                      .description   = "value",
                                      .units         = "",
                                      .value_type    = CLX_API_DATA_TYPE_UINT64,
                                      .counting_mode = CLX_FIELD_VALUE_ABSOLUTE,
                                      .offset        = 0,  // ignored
                                      .length        = sizeof(uint64_t)};

    uint32_t counter_num = 0;
    for (uint16_t i = 0; i < num_counters; i++) {
        if (asprintf(&counter.counter_name, "name_%u", i) < 0) {
            puts("Error: failed to construct a counter name");
            return false;
        }
        bool ok = clx_api_add_counter(ctx, &counter, group_name, &counter_num);
        if (i == 0) {
            *first_counter_idx = counter_num;
        }
        counter.counter_name = NULL;
        if (!ok) {
            puts("Error: failed to add counter");
            free(counter.counter_name);
            return false;
        }
        printf("    added counter #%u to group %s, returned number %u\n", i, group_name, counter_num);
    }

    return true;
}

bool setup_example_counters(clx_api_context_t* ctx, example_provider_context_t* context) {
    char* group_name = NULL;
    for (uint16_t i = 0; i < context->num_groups; i++) {
        if (asprintf(&group_name, "group_%u", i) < 0) {
            puts("Error: failed to construct a counter name");
            return false;
        }
        bool ok = setup_example_group_counters(ctx, group_name, context->num_counters, &(context->first_counter_idx[i]));
        free(group_name);
        group_name = NULL;
        if (!ok) {
            puts("Error: failed to add group counters");
            return false;
        }
    }
    return true;
}

bool initialize_example_provider(clx_api_context_t* ctx, clx_api_provider_t* provider) {
    puts("Initializing the data provider...");

    example_provider_context_t* context = (example_provider_context_t*)provider->data;

    if (!setup_example_counters(ctx, context)) {
        puts("Error: failed to construct counters.");
        return false;
    }

    return true;
}

void finalize_example_provider(clx_api_context_t* ctx, clx_api_provider_t* provider) { puts("Finalizing the data provider..."); }

void destroy_example_provider(clx_api_provider_t* provider) {
    if (provider != NULL) {
        free(provider->name);
        provider->data = NULL;
    }
    free(provider);
}

clx_api_provider_t* construct_example_provider(example_provider_context_t* provider_context) {
    clx_api_provider_t* provider = calloc(1, sizeof(clx_api_provider_t));
    if (provider == NULL) {
        return NULL;
    }
    clx_api_version_t ver = {{1, 0, 0}};
    provider->version     = ver;
    provider->name        = strdup("sink_test_provider");
    if (provider->name == NULL) {
        goto error;
    }

    provider->initialize = &initialize_example_provider;
    provider->finalize   = &finalize_example_provider;

    provider->data = provider_context;

    return provider;

error:
    destroy_example_provider(provider);
    return NULL;
}

//
// Main execution context
//
int main(void) {
    int ret                                      = 1;
    example_provider_context_t* provider_context = NULL;
    clx_api_provider_t* provider                 = NULL;
    clx_api_params_t* params                     = NULL;
    clx_api_context_t* api_context               = NULL;

    provider_context = construct_example_provider_context(kNumberOfCounterGroups, kNumberOfCounters);
    if (provider_context == NULL) {
        goto exit;
    }

    provider = construct_example_provider(provider_context);
    if (provider == NULL) {
        printf("Error: failed to initialize data provider.");
        goto exit;
    }

    params = clx_api_params_create();
    if (params == NULL) {
        puts("Error: unable to create parameters object");
        goto exit;
    }

    if (!set_example_parameters(params)) {
        puts("Error: unable to initialize parameters");
        goto exit;
    }

    print_example_parameters(params, stdout);

    api_context = clx_api_create_context(params, provider);
    if (api_context == NULL) {
        printf("Error: failed to create CLX API context.");
        goto exit;
    }

    uint64_t last_drop_timestamp = clx_api_get_timestamp();

    puts("Running the sampling loop...");

    for (uint32_t i = 0; i < kNumberOfSamples; i++) {
        uint64_t timestamp = clx_api_get_timestamp();

        // force page drop every X milliseconds
        if ((timestamp - last_drop_timestamp) / 1000 >= kForcePageWritePeriod) {
            puts("Forcing data transfer...");
            clx_api_force_write(api_context);
            last_drop_timestamp = timestamp;
        }

        // Sample counters
        if (!sample_counters(provider_context, api_context)) {
            printf("Error: counters sampling failed.");
            goto exit;
        }

        // Post bulk data once in a while
        if (i % 3 == 0) {
            if (!write_bulk_data(provider_context, api_context)) {
                printf("Error: bulk data writing failed.");
                goto exit;
            }
        }

        usleep(kSamplingPeriod * 1000);
    }

    ret = 0;  // indicate success

exit:
    clx_api_destroy_context(api_context);
    clx_api_params_delete(params);
    destroy_example_provider(provider);
    destroy_example_provider_context(provider_context);

    return ret;
}
